# BSD 2-Clause License

# Copyright 2022, Quoc-Tung Le

# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation 
#     and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.

import utils
import hierarchical_fact as fact
from scipy.linalg import hadamard

from iterative_method import training
import matplotlib.pyplot as plt
import matplotlib

import pandas as pd
import pickle
import time

import numpy as np


if __name__ == '__main__':
    num_factors = 10
    size = 2 ** num_factors
    support1 = utils.partial_prod_butterfly_supports(num_factors, 0, num_factors//2)
    support2 = utils.partial_prod_butterfly_supports(num_factors, num_factors//2, num_factors)
    learning_rate =[5e-1, 1e-1, 5e-2, 1e-2, 5e-3, 1e-3, 5e-4, 1e-4] 

    optimize = ['GD', 'GDMomentum', 'Adam']
    matrix = hadamard(size)
    speed_dict = dict()
    loss_dict = dict()
    
    begin = time.time()
    ft1, ft2 = fact.solve_DTO(support1, support2, matrix, type='real')
    end = time.time()

    speed_dict['SVD'] = end - begin
    loss_dict['SVD'] = np.linalg.norm(ft1 @ ft2 - matrix)
    
    for method in optimize:
        best_time = np.inf
        best_error = np.inf
        for lr in learning_rate:
            success, loss_array, time_array = training(matrix, support1, support2, lr = lr, option = method)
            if not success:
                continue
            if (loss_array[-1] < 1e-10 or loss_array[-1] < best_error) and time_array[-1] < best_time: 
                best_error = loss_array[-1]
                best_time = time_array[-1]    
                speed_dict[method] = time_array
                loss_dict[method] = np.log10(np.array(loss_array) + 1e-15)

    support1 = utils.partial_prod_butterfly_supports(num_factors, 0, 1)
    support2 = utils.partial_prod_butterfly_supports(num_factors, 1, num_factors)
    
    speed_dict2 = dict()
    loss_dict2 = dict()

    
    begin = time.time()
    ft1, ft2 = fact.solve_DTO(support1, support2, matrix, type='real')
    end = time.time()

    speed_dict2['SVD'] = end - begin
    loss_dict2['SVD'] = np.linalg.norm(ft1 @ ft2 - matrix)

    for method in optimize:
        best_time = np.inf
        best_error = np.inf
        for lr in learning_rate:
            success, loss_array, time_array = training(matrix, support1, support2, lr = lr, option = method)
            if not success:
                continue
            if (loss_array[-1] < 1e-10 or loss_array[-1] < best_error) and time_array[-1] < best_time: 
                best_error = loss_array[-1]
                best_time = time_array[-1]    
                speed_dict2[method] = time_array
                loss_dict2[method] = np.log10(np.array(loss_array) + 1e-15)

    with open("./data/speed_balance.pkl", 'wb') as handle:
        pickle.dump(speed_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)
    with open("./data/loss_balance.pkl", 'wb') as handle:
        pickle.dump(loss_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)
    with open("./data/speed_imbalance.pkl", 'wb') as handle:
        pickle.dump(speed_dict2, handle, protocol=pickle.HIGHEST_PROTOCOL)
    with open("./data/loss_imbalance.pkl", 'wb') as handle:
        pickle.dump(loss_dict2, handle, protocol=pickle.HIGHEST_PROTOCOL)
    

    with open("./data/speed_balance.pkl", 'rb') as handle:
        speed_dict = pickle.load(handle)
    with open("./data/loss_balance.pkl", 'rb') as handle:
        loss_dict = pickle.load(handle)
    with open("./data/speed_imbalance.pkl", 'rb') as handle:
        speed_dict2 = pickle.load(handle)
    with open("./data/loss_imbalance.pkl", 'rb') as handle:
        loss_dict2 = pickle.load(handle)

    fontsize = 15
    
    fig, (ax1, ax2) = plt.subplots(1,2, figsize = (15, 5))
    for method in optimize:
        ax1.plot(speed_dict[method], loss_dict[method], label=method)
    ax1.scatter(speed_dict['SVD'], np.log10(loss_dict['SVD']), label='Algorithm 3.4', color = 'red', marker = '*', s = 80)
    ax1.tick_params(axis='both', which='major', labelsize=fontsize)
    ax1.set_xlabel("running time (s)", fontsize = fontsize)
    ax1.set_ylabel(r'$\log_{10} \|A - XY^\top\|_F$', fontsize = fontsize)
    ax1.set_yticks([-12, -10, -8, -6, -4, -2, 0, 2])
    ax1.set_title("a)", fontsize = fontsize)
    ax1.legend()
    ax1.grid()

    for method in optimize:
        ax2.plot(speed_dict2[method], loss_dict2[method], label=method)
    ax2.scatter(speed_dict2['SVD'], np.log10(loss_dict2['SVD']), label='Algorithm 3.4', color = 'red', marker = '*', s = 80)
    ax2.tick_params(axis='both', which='major', labelsize=fontsize)
    ax2.set_xlabel("running time (s)", fontsize = fontsize)
    ax2.set_title("b)", fontsize = fontsize)
    ax2.set_yticks([-12, -10, -8, -6, -4, -2, 0, 2])
    ax2.legend()
    ax2.grid()
    
    plt.show()
